#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "easel.h"
#include "esl_alphabet.h"
#include "esl_getopts.h"
#include "esl_sq.h"
#include "esl_sqio.h"
#include "esl_vectorops.h"


static ESL_OPTIONS options[] = {
  /* name          type       default  env  range toggles reqs incomp  help         docgroup */
 { "-h",          eslARG_NONE,  FALSE, NULL, NULL, NULL, NULL, NULL, "show brief help",    0 },
 { "--dna",       eslARG_NONE,  FALSE, NULL, NULL, NULL, NULL, NULL, "use DNA alphabet",   0 },
 { "--rna",       eslARG_NONE,  FALSE, NULL, NULL, NULL, NULL, NULL, "use RNA alphabet",   0 },
 { "--amino",     eslARG_NONE,  FALSE, NULL, NULL, NULL, NULL, NULL, "use amino alphabet", 0 },
 {  0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
};
static char usage[]  = "[-options] <seqfile>";
static char banner[] = "test";

static int read_sq(ESL_ALPHABET *abc, ESL_SQFILE *sqfp);

int
main(int argc, char **argv)
{ 
  ESL_GETOPTS    *go = esl_getopts_CreateDefaultApp(options, 1, argc, argv, banner, usage);
  char           *infile = esl_opt_GetArg(go, 1);
  ESL_ALPHABET   *abc = NULL;
  ESL_SQFILE     *sqfp = NULL;
  int             alphatype = eslUNKNOWN;
  int             status;

  status = esl_sqfile_Open(infile, eslSQFILE_UNKNOWN, NULL, &sqfp);
  if      (status == eslENOTFOUND) esl_fatal("No such file.");
  else if (status == eslEFORMAT)   esl_fatal("Format couldn't be determined.");
  else if (status != eslOK)        esl_fatal("Open failed, code %d.", status);

  if      (esl_opt_GetBoolean(go, "--rna"))   alphatype = eslRNA;
  else if (esl_opt_GetBoolean(go, "--dna"))   alphatype = eslDNA;
  else if (esl_opt_GetBoolean(go, "--amino")) alphatype = eslAMINO;
  else {
    status = esl_sqfile_GuessAlphabet(sqfp, &alphatype);
    if      (status == eslENOALPHABET)  esl_fatal("Couldn't guess alphabet");
    else if (status == eslEFORMAT)      esl_fatal("Parse failed\n  %s",
						  esl_sqfile_GetErrorBuf(sqfp));     
    else if (status == eslENODATA)      esl_fatal("Sequence file empty?");
    else if (status != eslOK)           esl_fatal("Unexpected error guessing alphabet");
  }
  abc = esl_alphabet_Create(alphatype);

  status = read_sq(abc, sqfp);
  if (status != eslOK)  esl_fatal("Failed to run foo");


 esl_sqfile_Close(sqfp);
  esl_alphabet_Destroy(abc);
  esl_getopts_Destroy(go);
  return 0;
}


static
int read_sq(ESL_ALPHABET *abc, ESL_SQFILE *sqfp)
{
  ESL_SQ *s = NULL;
  int     n = 0;
  int     status;
  
  s = esl_sq_CreateDigital(abc);
  while ((status = esl_sqio_Read(sqfp, s)) == eslOK) {
    esl_sqio_Write(stdout, s, eslSQFILE_FASTA, /*update=*/FALSE);
    
    esl_sq_Reuse(s);
    n++;
  }
  if (status != eslEOF) goto ERROR;
  printf("N %d\n", n);
  
  free(s);
  return eslOK;

 ERROR:
  if (s) esl_sq_Destroy(s);
  return status;
}



